"""LIKE statement"""

from typing import Any, NoReturn

from sqlfactory.condition.base import ConditionBase, StatementOrColumn
from sqlfactory.entities import Column
from sqlfactory.statement import Statement


class Like(ConditionBase):
    """
    SQL `LIKE` condition for comparing strings against pattern.

    Examples:

    - Simple
        ```python
        # `column` LIKE %s
        Like("column", "pattern")
        "`column` LIKE %s", ["pattern"]
        ```
    - Negative
        ```python
        # `column` NOT LIKE %s
        Like("column", "pattern", negative=True)
        "`column` NOT LIKE %s", ["pattern"]
        ```
    - Statement
        ```python
        Like("column", Concat("%", Column("other_column"), "%"))
        "`column` LIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Statement (negative)
        ```python
        Like("column", Concat("%", Column("other_column"), "%"), negative=True)
        "`column` NOT LIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Instead of column, you can also use any other expression
        ```python
        Like(Concat("column", "other_column"), "pattern")
        "CONCAT(`column`, `other_column`) LIKE %s", ["pattern"]
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement, negative: bool = False) -> None:
        """
        :param column: Column (or statement) on left side of LIKE operator.
        :param value: Value to match the column (or statement) against.
        :param negative: Whether to use negative matching (NOT LIKE).
        """

        super().__init__()

        if not isinstance(column, Statement):
            column = Column(column)

        self._column = column
        self._value = value
        self._negative = negative

    @staticmethod
    def escape(s: str) -> str:
        """
        Escape string for use in LIKE statement
        :param s: String to be escaped
        :return: String with escaped characters - % -> %%, _ -> __, to be safely used as part of pattern in LIKE statement.
        """
        return s.replace("%", "%%").replace("_", "__")

    def __str__(self) -> str:
        if isinstance(self._value, Statement):
            return f"{self._column!s}{' NOT' if self._negative else ''} LIKE {self._value!s}"

        return f"{self._column!s}{' NOT' if self._negative else ''} LIKE {self.dialect.placeholder}"

    @property
    def args(self) -> list[Any]:
        """Like statement arguments."""

        args = [*self._column.args]

        if isinstance(self._value, Statement):
            args.extend(self._value.args)
        else:
            args.append(self._value)

        return args

    def __bool__(self) -> bool:
        return True

    def __invert__(self) -> "Like":
        """
        Allows using the `~` operator to negate the LIKE condition, converting it to a NOT LIKE condition.
        Note: Cannot use ~ operator on NotLike conditions.
        """
        return NotLike(self._column, self._value)


class NotLike(Like):
    """
    SQL `NOT LIKE` condition for comparing strings against pattern.

    This is a dedicated class for NOT LIKE conditions, which is equivalent to using Like with negative=True.
    It provides a more intuitive API for NOT LIKE conditions.

    Examples:

    - Simple
        ```python
        # `column` NOT LIKE %s
        NotLike("column", "pattern")
        "`column` NOT LIKE %s", ["pattern"]
        ```
    - Statement
        ```python
        NotLike("column", Concat("%", Column("other_column"), "%"))
        "`column` NOT LIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Instead of column, you can also use any other expression
        ```python
        NotLike(Concat("column", "other_column"), "pattern")
        "CONCAT(`column`, `other_column`) NOT LIKE %s", ["pattern"]
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column (or statement) on left side of NOT LIKE operator.
        :param value: Value to match the column (or statement) against.
        """
        super().__init__(column, value, negative=True)

    def __invert__(self) -> NoReturn:
        """
        Allows using the `~` operator to negate the LIKE condition, converting it to a NOT LIKE condition.
        Note: Cannot use ~ operator on NotLike conditions.
        """
        raise TypeError("Cannot use ~ operator on NotLike conditions")



"""Encryption functions (https://mariadb.com/kb/en/encryption-hashing-and-compression-functions/)"""

from typing import Any, Literal, Optional

from sqlfactory.func.base import Function
from sqlfactory.statement import Statement


class AesDecrypt(Function):
    """
    Decrypts an encrypted string using AES.
    """

    def __init__(self, value: Any, key: Any):
        super().__init__("AES_DECRYPT", value, key)


class AesEncrypt(Function):
    """
    Encrypts a string using AES.
    """

    def __init__(self, value: Any, key: Any):
        super().__init__("AES_ENCRYPT", value, key)


class Compress(Function):
    """
    Compress a string.
    """

    def __init__(self, value: Any):
        super().__init__("COMPRESS", value)


class DesDecrypt(Function):
    """
    Decrypts an encrypted string using DES.
    """

    def __init__(self, value: Any, key: Any):
        super().__init__("DES_DECRYPT", value, key)


class DesEncrypt(Function):
    """
    Encrypts a string using DES.
    """

    def __init__(self, value: Any, key: Any):
        super().__init__("DES_ENCRYPT", value, key)


class Encode(Function):
    """
    Encode a string.
    """

    def __init__(self, value: Any, encoding: str):
        super().__init__("ENCODE", value, encoding)


class Decode(Function):
    """
    Decode a string.
    """

    def __init__(self, value: Any, encoding: str):
        super().__init__("DECODE", value, encoding)


class Encrypt(Function):
    """
    Encrypt a string using Unix crypt()
    """

    def __init__(self, value: Any, salt: Any = None):
        if salt is None:
            super().__init__("ENCRYPT", value)
        else:
            super().__init__("ENCRYPT", value, salt)


class Kdf(Function):
    """
    Key derivation function.
    """

    def __init__(
        self,
        key: Any,
        salt: Any,
        info_or_iterations: Any = None,
        kdf_name: Optional[Literal["hkdf", "pbkdf2_hmac"] | Statement] = None,
        width: Any = None,
    ):
        args = [key, salt]

        if info_or_iterations is not None:
            args.append(info_or_iterations)

            if kdf_name is not None:
                args.append(kdf_name)

                if width is not None:
                    args.append(width)

        super().__init__("KDF", *args)


class OldPassword(Function):
    """
    Hash a string using the old MySQL password hashing algorithm.
    """

    def __init__(self, value: Any):
        super().__init__("OLD_PASSWORD", value)


class Password(Function):
    """
    Hash a string using the MySQL password hashing algorithm.
    """

    def __init__(self, value: Any):
        super().__init__("PASSWORD", value)


class MD5(Function):
    """
    Calculate an MD5 128-bit checksum.
    """

    def __init__(self, value: Any):
        super().__init__("MD5", value)


class RandomBytes(Function):
    """
    Generate a random byte string.
    """

    def __init__(self, length: Any):
        super().__init__("RANDOM_BYTES", length)


class Sha1(Function):
    """
    Calculate an SHA-1 160-bit checksum.
    """

    def __init__(self, value: Any):
        super().__init__("SHA1", value)


class Sha2(Function):
    """
    Calculate an SHA-2 checksum.
    """

    def __init__(self, value: Any, length: Any):
        super().__init__("SHA2", value, length)


class Uncompress(Function):
    """
    Uncompress a compressed string.
    """

    def __init__(self, value: Any):
        super().__init__("UNCOMPRESS", value)


class UncompressLength(Function):
    """
    Return the length of an uncompressed string.
    """

    def __init__(self, value: Any):
        super().__init__("UNCOMPRESS_LENGTH", value)



from typing import Any, Self, TypeAlias

from sqlfactory.dialect import SQLDialect
from sqlfactory.execute import ExecutableStatement
from sqlfactory.mixins.limit import Limit, WithLimit
from sqlfactory.mixins.order import OrderArg, WithOrder
from sqlfactory.select.select import Select


class Union(ExecutableStatement, WithOrder, WithLimit):
    """
    Construct UNION statement by combining multiple SELECTs.

    Example:

    >>> from sqlfactory import Select, Union
    >>> sel = Union(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) UNION (SELECT a, b FROM table2)

    Optionally you can specify ordering and limit:

    >>> from sqlfactory import Select, Union, Direction, Limit
    >>> sel = Union(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ...     order=[("a", Direction.ASC)],
    ...     limit=Limit(10),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) UNION (SELECT a, b FROM table2) ORDER BY a ASC LIMIT %s
    >>> print(sel.args)
    [10]

    The same options for ordering and limit apply as for `Select` statement.

    """

    def __init__(
        self,
        *selects: Select,
        order: OrderArg | None = None,
        limit: Limit | None = None,
        dialect: SQLDialect | None = None,
    ) -> None:
        """
        Construct UNION statement by combining multiple SELECTs.

        :param selects: SELECTs to be combined
        :param order: Ordering specification (same as for `Select`)
        :param limit: Limit specification (same as for `Select`)
        """
        super().__init__(order=order, limit=limit, dialect=dialect)
        self._selects = list(selects)

    def append(self, select: Select) -> Self:
        """
        Append another SELECT to the UNION.
        :param select: Select to be appended
        :return: Self for chaining
        """
        self._selects.append(select)
        return self

    @property
    def _joiner(self) -> str:
        return " UNION "

    def __str__(self) -> str:
        out = [self._joiner.join(f"({select!s})" for select in self._selects)]

        if self._order:
            out.append(str(self._order))

        if self._limit:
            out.append(str(self._limit))

        return " ".join(out)

    @property
    def args(self) -> list[Any]:
        """
        Argument values for the UNION statement and all sub-statements.
        """
        out: list[Any] = []

        for select in self._selects:
            out.extend(select.args)

        if self._order:
            out.extend(self._order.args)

        if self._limit:
            out.extend(self._limit.args)

        return out

    def __bool__(self) -> bool:
        return bool(self._selects)


class UnionAll(Union):
    """
    The same as `Union`, but uses `UNION ALL` instead of `UNION` to join SELECTs.

    Example:

    >>> from sqlfactory import Select, UnionAll
    >>> sel = UnionAll(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) UNION ALL (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " UNION ALL "


class UnionDistinct(Union):
    """
    The same as `Union`, but uses `UNION DISTINCT` instead of `UNION` to join SELECTs. UNION DISTINCT is in fact alias for
    plain UNION, but it is provided for clarity if you want to be explicit about the DISTINCT keyword.

    Example:

    >>> from sqlfactory import Select, UnionDistinct
    >>> sel = UnionDistinct(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) UNION DISTINCT (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " UNION DISTINCT "


UNION: TypeAlias = Union  # pylint: disable=invalid-name
"""
Alias for `Union` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

UNION_ALL: TypeAlias = UnionAll  # pylint: disable=invalid-name
"""
Alias for `UnionAll` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

UNION_DISTINCT: TypeAlias = UnionDistinct  # pylint: disable=invalid-name
"""
Alias for `UnionDistinct` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""


class Except(Union):
    """
    Construct EXCEPT statement by combining multiple SELECTs.

    Example:

    >>> from sqlfactory import Select, Except
    >>> sel = Except(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) EXCEPT (SELECT a, b FROM table2)

    Optionally you can specify ordering and limit:

    >>> from sqlfactory import Select, Except, Direction, Limit
    >>> sel = Except(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ...     order=[("a", Direction.ASC)],
    ...     limit=Limit(10),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) EXCEPT (SELECT a, b FROM table2) ORDER BY a ASC LIMIT %s
    >>> print(sel.args)
    [10]

    """

    @property
    def _joiner(self) -> str:
        return " EXCEPT "


class ExceptAll(Except):
    """
    The same as `Except`, but uses `EXCEPT ALL` instead of `EXCEPT` to join SELECTs.

    Example:

    >>> from sqlfactory import Select, ExceptAll
    >>> sel = ExceptAll(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) EXCEPT ALL (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " EXCEPT ALL "


class ExceptDistinct(Except):
    """
    The same as `Except`, but uses `EXCEPT DISTINCT` instead of `EXCEPT` to join SELECTs. EXCEPT DISTINCT is in fact alias for
    plain EXCEPT, but it is provided for clarity if you want to be explicit about the DISTINCT keyword.

    Example:

    >>> from sqlfactory import Select, ExceptDistinct
    >>> sel = ExceptDistinct(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) EXCEPT DISTINCT (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " EXCEPT DISTINCT "


EXCEPT: TypeAlias = Except  # pylint: disable=invalid-name
"""
Alias for `Except` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

EXCEPT_ALL: TypeAlias = ExceptAll  # pylint: disable=invalid-name
"""
Alias for `ExceptAll` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

EXCEPT_DISTINCT: TypeAlias = ExceptDistinct  # pylint: disable=invalid-name
"""
Alias for `ExceptDistinct` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""


class Intersect(Union):
    """
    Construct INTERSECT statement by combining multiple SELECTs.

    Example:

    >>> from sqlfactory import Select, Intersect
    >>> sel = Intersect(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) INTERSECT (SELECT a, b FROM table2)

    Optionally you can specify ordering and limit:

    >>> from sqlfactory import Select, Intersect, Direction, Limit
    >>> sel = Intersect(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ...     order=[("a", Direction.ASC)],
    ...     limit=Limit(10),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) INTERSECT (SELECT a, b FROM table2) ORDER BY a ASC LIMIT %s
    >>> print(sel.args)
    [10]

    """

    @property
    def _joiner(self) -> str:
        return " INTERSECT "


class IntersectAll(Intersect):
    """
    The same as `Intersect`, but uses `INTERSECT ALL` instead of `INTERSECT` to join SELECTs.

    Example:

    >>> from sqlfactory import Select, IntersectAll
    >>> sel = IntersectAll(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) INTERSECT ALL (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " INTERSECT ALL "


class IntersectDistinct(Intersect):
    """
    The same as `Intersect`, but uses `INTERSECT DISTINCT` instead of `INTERSECT` to join SELECTs. INTERSECT DISTINCT is in fact
    alias for plain INTERSECT, but it is provided for clarity if you want to be explicit about the DISTINCT keyword.

    Example:

    >>> from sqlfactory import Select, IntersectDistinct
    >>> sel = IntersectDistinct(
    ...     Select("a", "b", table="table1"),
    ...     Select("a", "b", table="table2"),
    ... )
    >>> print(sel)
    (SELECT a, b FROM table1) INTERSECT DISTINCT (SELECT a, b FROM table2)

    """

    @property
    def _joiner(self) -> str:
        return " INTERSECT DISTINCT "


INTERSECT: TypeAlias = Intersect  # pylint: disable=invalid-name
"""
Alias for `Intersect` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

INTERSECT_ALL: TypeAlias = IntersectAll  # pylint: disable=invalid-name
"""
Alias for `IntersectAll` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""

INTERSECT_DISTINCT: TypeAlias = IntersectDistinct  # pylint: disable=invalid-name
"""
Alias for `IntersectDistinct` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""



"""RLIKE statement"""

from typing import Any, NoReturn

from sqlfactory.condition.base import StatementOrColumn
from sqlfactory.condition.like import Like
from sqlfactory.statement import Statement


class RLike(Like):
    """
    SQL `RLIKE` condition performs a pattern match of a string expression against a pattern.

    Examples:

    - Simple
        ```python
        # `column` RLIKE %s
        RLike("column", "pattern")
        "`column` RLIKE %s", ["pattern"]
        ```
    - Negative
        ```python
        # `column` NOT RLIKE %s
        RLike("column", "pattern", negative=True)
        "`column` NOT RLIKE %s", ["pattern"]
        ```
    - Statement
        ```python
        RLike("column", Concat(".*", Column("other_column"), ".*"))
        "`column` RLIKE CONCAT(%s, `other_column`, %s)", [".*", ".*"]
        ```
    - Statement (negative)
        ```python
        RLike("column", Concat(".*", Column("other_column"), ".*"), negative=True)
        "`column` NOT RLIKE CONCAT(%s, `other_column`, %s)", [".*", ".*"]
        ```
    - Instead of column, you can also use any other expression
        ```python
        RLike(Concat("column", "other_column"), "pattern")
        "CONCAT(`column`, `other_column`) RLIKE %s", ["pattern"]
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement, negative: bool = False) -> None:
        """
        :param column: Column (or statement) on left side of RLIKE operator.
        :param value: Value to match the column (or statement) against.
        :param negative: Whether to use negative matching (NOT RLIKE).
        """
        # we want change init comment to be more specific
        # pylint: disable=useless-parent-delegation
        super().__init__(column, value, negative)

    def __str__(self) -> str:
        if isinstance(self._value, Statement):
            return f"{self._column!s}{' NOT' if self._negative else ''} RLIKE {self._value!s}"

        return f"{self._column!s}{' NOT' if self._negative else ''} RLIKE {self.dialect.placeholder}"

    def __invert__(self) -> "RLike":
        """
        Allows using the `~` operator to negate the RLIKE condition, converting it to a NOT RLIKE condition.
        Note: Cannot use ~ operator on NotRLike conditions.
        """
        return NotRLike(self._column, self._value)


class NotRLike(RLike):
    """
    SQL `NOT RLIKE` condition performs a pattern match of a string expression against a pattern.

    This is a dedicated class for NOT RLIKE conditions, which is equivalent to using RLike with negative=True.
    It provides a more intuitive API for NOT RLIKE conditions.

    Examples:

    - Simple
        ```python
        # `column` NOT RLIKE %s
        NotRLike("column", "pattern")
        "`column` NOT RLIKE %s", ["pattern"]
        ```
    - Statement
        ```python
        NotRLike("column", Concat("%", Column("other_column"), "%"))
        "`column` NOT RLIKE CONCAT(%s, `other_column`, %s)", ["%", "%"]
        ```
    - Instead of column, you can also use any other expression
        ```python
        NotRLike(Concat("column", "other_column"), "pattern")
        "CONCAT(`column`, `other_column`) NOT RLIKE %s", ["pattern"]
        ```
    """

    def __init__(self, column: StatementOrColumn, value: Any | Statement) -> None:
        """
        :param column: Column (or statement) on left side of NOT RLIKE operator.
        :param value: Value to match the column (or statement) against.
        """
        super().__init__(column, value, negative=True)

    def __invert__(self) -> NoReturn:
        """
        Allows using the `~` operator to negate the RLIKE condition, converting it to a NOT RLIKE condition.
        Note: Cannot use ~ operator on NotRLike conditions.
        """
        raise TypeError("Cannot use ~ operator on NotRLike conditions")



"""Information functions (https://mariadb.com/kb/en/information-functions/)."""

from typing import Any

from sqlfactory.entities import Column
from sqlfactory.func.base import Function
from sqlfactory.statement import Statement


class Benchmark(Function):
    """Executes an expression repeatedly."""

    def __init__(self, count: int, expression: Statement) -> None:
        super().__init__("BENCHMARK", count, expression)


class BinlogGtidPos(Function):
    """Returns a string representation of the corresponding GTID position."""

    def __init__(self) -> None:
        super().__init__("BINLOG_GTID_POS")


class Charset(Function):
    """Returns the character set."""

    def __init__(self) -> None:
        super().__init__("CHARSET")


class Coercibility(Function):
    """Returns the collation coercibility value of the string expression."""

    def __init__(self, expression: str) -> None:
        super().__init__("COERCIBILITY", expression)


class Collation(Function):
    """Collation of the string argument"""

    def __init__(self, expression: str) -> None:
        super().__init__("COLLATION", expression)


class Collate(Statement):
    """String with collation"""

    def __init__(self, expression: str | Statement, collation: str) -> None:
        super().__init__()

        self._expression = expression
        self._collation = collation

    def __str__(self) -> str:
        return (
            f"{str(self._expression) if isinstance(self._expression, Statement) else self.dialect.placeholder} "
            f"COLLATE {self._collation}"
        )

    @property
    def args(self) -> list[Any]:
        return self._expression.args if isinstance(self._expression, Statement) else [self._expression]


class ConnectionId(Function):
    """Connection ID"""

    def __init__(self) -> None:
        super().__init__("CONNECTION_ID")


class CurrentRole(Function):
    """Current role name"""

    def __init__(self) -> None:
        super().__init__("CURRENT_ROLE")


class CurrentUser(Function):
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("CURRENT_USER")


class Database(Function):
    """Current default database"""

    def __init__(self) -> None:
        super().__init__("DATABASE")


class DecodeHistogram(Function):
    """Returns comma separated numerics corresponding to a probability distribution"""

    def __init__(self, hist_type: Any, histogram: Any) -> None:
        super().__init__("DECODE_HISTOGRAM", hist_type, histogram)


class Default(Function):
    """Returns the default value for a table column"""

    def __init__(self, column: Column) -> None:
        super().__init__("DEFAULT", column)


class FoundRows(Function):
    """Returns the number of (potentially) returned rows if there was no LIMIT involved."""

    def __init__(self) -> None:
        super().__init__("FOUND_ROWS")


class LastInsertId(Function):
    """Returns the value generated for an AUTO_INCREMENT column by the previous INSERT statement."""

    def __init__(self) -> None:
        super().__init__("LAST_INSERT_ID")


class LastValue(Function):
    """Evaluates expression and returns the last."""

    def __init__(self, expr: Statement, *exprs: Statement) -> None:
        super().__init__("LAST_VALUE", expr, *exprs)


class RowNumber(Function):
    """Returns the number of accepted rows so far."""

    def __init__(self) -> None:
        super().__init__("ROW_NUMBER")


class Schema(Function):
    """Current default schema"""

    def __init__(self) -> None:
        super().__init__("SCHEMA")


class SessionUser(Function):
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("SESSION_USER")


class SystemUser(Function):
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("SYSTEM_USER")


class User(Function):
    """Username/host that authenticated the current client"""

    def __init__(self) -> None:
        super().__init__("USER")


class Version(Function):
    """Database version"""

    def __init__(self) -> None:
        super().__init__("VERSION")



# noqa: A005

"""SELECT statement builder."""

from __future__ import annotations

from collections.abc import Collection
from functools import reduce
from typing import Any, Self, TypeAlias, cast

from sqlfactory.condition.base import And, ConditionBase
from sqlfactory.dialect import SQLDialect
from sqlfactory.entities import Column, ColumnArg, Table
from sqlfactory.execute import ExecutableStatement
from sqlfactory.mixins.join import WithJoin
from sqlfactory.mixins.limit import Limit, WithLimit
from sqlfactory.mixins.order import OrderArg, WithOrder
from sqlfactory.mixins.where import WithWhere
from sqlfactory.select.column_list import ColumnList
from sqlfactory.select.join import Join
from sqlfactory.statement import Statement


class Select(ExecutableStatement, WithWhere, WithOrder, WithLimit, WithJoin):
    # pylint: disable=too-many-arguments  # Yes, SELECT is complex.
    """
    `SELECT` statement to create complex select queries.

    Example:

    >>> from sqlfactory import Eq
    >>> cursor: Cursor = ...
    >>>
    >>> (
    ...     Select("column1", "column2", "column3", table="table_name")
    ...     .where(Eq("column1", 1))
    ...     .order_by("column2")
    ...     .limit(2, 10)
    ...     .execute(cursor)
    ... )

    **Note:** `Eq()` and all other condition functions are expecting first argument to be column and second argument to be value.
    If you want to have column on the right side or value on the left, you must explicitely use `Column()` function. And if you
    want literal value on the left side, you must explicitely use `Value()` function.

    Subqueries are also supported:

    >>> from sqlfactory import Select, Aliased
    >>> from sqlfactory.func.agg import Sum
    >>>
    >>> Select(
    ...     "t1.id", "t2.price",
    ...     table=[
    ...         "t1",
    ...         Aliased(Select(Aliased(Sum("price"), alias="price"), table="product_prices"), alias="t2")
    ...     ]
    ... )

    Known limitations:
    - JOINs are not checked for uniqueness, so it is possible to add same JOIN multiple times. For now, it is up to
      user to ensure that JOINs are unique. This may be changed in the future, but it is not simple task, as joins
      needs to be in certain order.
    """

    def __init__(
        self,
        *columns: Statement | ColumnArg,
        select: ColumnList | list[str] | None = None,
        table: Table | str | Statement | Collection[Table | str | Statement] | None = None,
        join: Collection[Join] | None = None,
        where: ConditionBase | None = None,
        group_by: ColumnList | Collection[Statement | ColumnArg] | None = None,
        having: ConditionBase | None = None,
        order: OrderArg | None = None,
        limit: Limit | None = None,
        for_update: bool = False,
        dialect: SQLDialect | None = None,
    ) -> None:
        """
        :param columns: Columns to select.
        :param select: Columns to select as instance of ColumnList instead of specifying positional arguments.
        :param table: Table to select from.
        :param join: Join statements
        :param where: Where condition
        :param group_by: Group by columns
        :param having: Having condition
        :param order: Order by columns
        :param limit: Limit results
        :param for_update: Lock rows for update
        """
        super().__init__(where=where, order=order, limit=limit, join=join, dialect=dialect)

        if columns and select:
            raise AttributeError("Cannot specify individual columns when attribute select is present.")

        if select and not isinstance(select, ColumnList):
            select = ColumnList(select)

        self.columns: ColumnList = cast(ColumnList | None, select) or ColumnList(columns)
        """Columns to select."""

        if (table is not None and not isinstance(table, Collection)) or isinstance(table, str):
            table = [table]

        self.table: list[Statement] = [Table(t) if isinstance(t, str) else t for t in table] if table is not None else []
        """Table (or tables) to select from."""

        self._group_by = ColumnList(group_by) if group_by is not None and not isinstance(group_by, ColumnList) else group_by
        self._having = having
        self._for_update = for_update

    def add(self, column: Statement | Any) -> Self:
        """Add new statement or column to the set of selected columns"""
        self.columns.add(column)
        return self

    def group_by(self, column: Statement | ColumnArg, *columns: Statement | ColumnArg) -> Self:
        """
        `GROUP BY` clause.

        >>> Select().group_by("column1", "column2", "column3")
        >>> "SELECT ... GROUP BY `column1`, `column2`, `column3`"

        **Note:** When calling `Select.group_by()` multiple times, it will join the grouped columns together.
        """
        if self._group_by is not None:
            self._group_by.append(column)

            if columns:
                self._group_by.extend([Column(column) if not isinstance(column, Statement) else column for column in columns])

        else:
            self._group_by = ColumnList([column, *list(columns)])

        return self

    # pylint: disable=invalid-name
    def GROUP_BY(self, column: Statement | ColumnArg, *columns: Statement | ColumnArg) -> Self:
        """Alias for `Select.group_by()` to be more SQL-like with all capitals."""
        return self.group_by(column, *columns)

    def having(self, condition: ConditionBase) -> Self:
        """
        `HAVING` clause

        >>> Select().having(Eq("column1", 3))
        >>> "SELECT ... HAVING `column1` = %s", [3]
        """
        if self._having is not None:
            if isinstance(self._having, And):
                self._having.append(condition)
            else:
                self._having = And(self._having, condition)
        else:
            self._having = condition

        return self

    # pylint: disable=invalid-name
    def HAVING(self, condition: ConditionBase) -> Self:
        """Alias for `Select.having()` to be more SQL-like with all capitals."""
        return self.having(condition)

    def __str__(self) -> str:
        with self.dialect:
            if not self.columns and not self.table:
                return ""

            out: list[str] = ["SELECT", str(self.columns) if self.columns else "*"]

            if self.table:
                out.append(f"FROM {', '.join(map(str, self.table))}")

            if self._join:
                out.extend(map(str, self._join))

            if self._where:
                out.append("WHERE")
                out.append(str(self._where))

            if self._group_by:
                out.append("GROUP BY")
                out.append(str(self._group_by))

            if self._having:
                out.append("HAVING")
                out.append(str(self._having))

            if self._order:
                out.append(str(self._order))

            if self._limit:
                out.append(str(self._limit))

            if self._for_update:
                out.append("FOR UPDATE")

            return " ".join(out)

    @property
    def args(self) -> list[Any]:
        """Argument values for the SELECT statement."""
        out = self.columns.args

        if self._join:
            for join in self._join:
                out.extend(join.args)

        return (
            out
            + reduce(lambda acc, t: acc + (t.args if isinstance(t, Statement) else []), self.table, [])
            + (self._where.args if self._where else [])
            + (self._group_by.args if self._group_by else [])
            + (self._having.args if self._having else [])
            + (self._order.args if self._order else [])
            + (self._limit.args if self._limit else [])
        )

    def __bool__(self) -> bool:
        return bool(self.columns) or bool(self.table)


SELECT: TypeAlias = Select  # pylint: disable=invalid-name
"""
Alias for `Select` statement to provide better SQL compatibility, as SQL is often written in all caps.
"""



"""IN condition, used for checking whether column value is in given list of values."""

from __future__ import annotations

from collections.abc import Collection
from typing import TYPE_CHECKING, Any, NoReturn, cast, overload

from sqlfactory.condition.base import And, ConditionBase, Or, StatementOrColumn
from sqlfactory.condition.simple import Eq, Ne
from sqlfactory.entities import Column
from sqlfactory.statement import Raw, Statement

if TYPE_CHECKING:
    from sqlfactory.select.select import Select  # pragma: no cover


class In(ConditionBase):
    """
    `IN` condition for checking whether column value is in given list of values.

    Supports both single-column and multi-column comparisons.

    In expects first argument to be column, and second argument to be list of values to compare against. So if you are
    passing strings, this is what happens. If you want to specify something else, you must be explicit. See examples.

    Examples:

    ### Single column

    - Simple IN statement:
        ```python
        # `column` IN (%s, %s, %s)
        In("column", [1, 2, 3])
        "`column` IN (%s, %s, %s)", [1, 2, 3]
        ```
    - Statement IN values
        ```python
        # <statement> IN (%s, %s, %s)
        In(Date("column"), ["2021-01-01", "2021-01-02", "2021-01-03"])
        "Date(`column`) IN (%s, %s, %s)", ["2021-01-01", "2021-01-02", "2021-01-03"]
        ```
    - Column IN columns
        ```python
        # Note the explicit `Column` usage.
        In("column", [Column("column1"), Column("column2"), Column("column3")])
        ```
    - Value with None works out-of-the-box
        ```python
        In("column", [1, None, 3])
        "`column` IN (%s, %s) OR `column` IS NULL", [1, 3]
        ```

    ### Multi column

    - Agains tuples of values:
        ```python
        # (`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))
        In(("column1", "column2"), [(1, 2), (3, 4), (5, 6)])
        "(`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))", [1, 2, 3, 4, 5, 6]
        ```
    - Mix of statement and column
        ```python
        # (<statement>, `column`) IN ((%s, %s), (%s, %s), (%s, %s))
        In((Date("column1"), "column2"), [("2021-01-01", 2), ("2021-01-02", 4), ("2021-01-03", 6)])
        "(DATE(`column1`), `column2`) IN ((%s, %s), (%s, %s), (%s, %s))", ["2021-01-01", 2, "2021-01-02", 4, "2021-01-03", 6]
        ```
    - Only statements
        ```python
        # (<statement>, <statement>) IN ((%s, %s), (%s, %s), (%s, %s))
        In((Min("column1"), Max("column1")), [(1, 2), (3, 4), (5, 6)])
        "(`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))", [1, 2, 3, 4, 5, 6]
        ```
    - Agains other columns or statements
        ```python
        In(("column1", Max("column2")), [(Min("column1"), 2), (3, Sum("column3")), (5, 6)])
        "(`column1`, MAX(`column2`)) IN ((MIN(`column1`), %s), (%s, SUM(`column3`)), (%s, %s))", [2, 3, 5, 6]
        ```
    - Values with `None`:
        ```python
        In(("column1", "column2"), [(1, 2), (3, None), (5, 6)])
        "(`column1`, `column2`) IN ((%s, %s), (%s, %s)) OR (`column1` = %s AND `column2` IS NULL)", [1, 2, 3, 5, 6]


    ### Subquery IN

    ```python
    In("column", Select("column", table="table", where=Eq("column", 1)))
    ```

    ```python
    In(("column1", "column2"), Select("column1", "column2", table="table", where=Eq("column", 1)))
    ```
    """

    @overload
    def __init__(
        self, columns: tuple[StatementOrColumn, ...], values: Collection[tuple[Any, ...]], /, negative: bool = False
    ) -> None:
        """Provides type definition for statement (`column1`, `column2`) IN ((%s, %s), (%s, %s), (%s, %s))"""

    @overload
    def __init__(self, columns: tuple[StatementOrColumn, ...], values: Select, /, negative: bool = False) -> None:
        """Provides type definition for statement (`column1`, `column2`) IN (SELECT ...)"""

    @overload
    def __init__(self, column: StatementOrColumn, values: Collection[Any], /, negative: bool = False) -> None:
        """Provides type definition for statement `column` IN (%s, %s, %s)"""

    @overload
    def __init__(self, column: StatementOrColumn, values: Select, /, negative: bool = False) -> None:
        """Provides type definition for statement `column` IN (SELECT ...)"""

    def __init__(
        self,
        column: StatementOrColumn | tuple[StatementOrColumn, ...],
        values: Collection[Any | tuple[Any, ...]] | Select,
        /,
        negative: bool = False,
    ) -> None:
        """
        :param column: Column to compare, or tuple of columns for multi-column comparison.
        :param values: Values to compare (list of values, or list of tuples of values for multi-column In).
        :param negative: Whether to perform negative comparison (NOT IN)
        """
        super().__init__()

        self._is_multi_column = isinstance(column, tuple)
        self._column = column
        self._values = values
        self._negative = negative

    def __str__(self) -> str:
        from sqlfactory.select.select import Select  # pylint: disable=import-outside-toplevel

        if isinstance(self._values, Select):
            stmt, _ = self._build_subquery_in(self._column, self._values, negative=self._negative)

        elif self._is_multi_column:
            stmt, _ = self._build_multi_in(cast(tuple[StatementOrColumn], self._column), self._values, negative=self._negative)
        else:
            stmt, _ = self._build_simple_in(cast(StatementOrColumn, self._column), self._values, negative=self._negative)

        return stmt

    @property
    def args(self) -> list[Any]:
        from sqlfactory.select.select import Select  # pylint: disable=import-outside-toplevel

        if isinstance(self._values, Select):
            _, args = self._build_subquery_in(self._column, self._values, negative=self._negative)

        elif self._is_multi_column:
            _, args = self._build_multi_in(cast(tuple[StatementOrColumn], self._column), self._values, negative=self._negative)
        else:
            _, args = self._build_simple_in(cast(StatementOrColumn, self._column), self._values, negative=self._negative)

        return args

    @staticmethod
    def _build_subquery_in(
        columns: StatementOrColumn | tuple[StatementOrColumn, ...], select: Select, *, negative: bool = False
    ) -> tuple[str, list[Any]]:
        # pylint: disable=consider-using-f-string
        args = []

        if isinstance(columns, tuple):
            in_columns = [Column(col) if not isinstance(col, Statement) else col for col in columns]

            for column in in_columns:
                args.extend(column.args)

            args.extend(select.args)

            in_stmt = "({}) {} ({})".format(
                ", ".join(map(str, in_columns)),
                "IN" if not negative else "NOT IN",
                str(select),
            )

        else:
            if not isinstance(columns, Statement):
                columns = Column(columns)

            args = [*columns.args, *select.args]
            in_stmt = "{} {} ({})".format(
                str(columns),
                "IN" if not negative else "NOT IN",
                str(select),
            )

        return in_stmt, args

    def _build_simple_in(
        self, column: StatementOrColumn, values: Collection[Any], *, negative: bool = False
    ) -> tuple[str, list[Any]]:
        # pylint: disable=consider-using-f-string
        if not isinstance(column, Statement):
            column = Column(column)

        add_none = any(value is None for value in values)
        if add_none:
            values = [value for value in values if value is not None]

        args = []

        if values:
            in_stmt = "{} {} ({})".format(
                str(column),
                "IN" if not negative else "NOT IN",
                ", ".join([self.dialect.placeholder if not isinstance(value, Statement) else str(value) for value in values]),
            )

            if isinstance(column, Statement):
                args.extend(column.args)

            for value in values:
                if isinstance(value, Statement):
                    args.extend(value.args)
                elif not isinstance(value, Statement):
                    args.append(value)

            if add_none:
                if isinstance(column, Statement):
                    args.extend(column.args)

                return (f"({in_stmt} {'OR' if not negative else 'AND'} {column!s} IS {'NOT ' if negative else ''}NULL)", args)

            return (in_stmt, args)

        if add_none:
            # This could happen only if there is just a one column, not multi-column statement.
            if isinstance(column, Statement):
                args.extend(column.args)

            return (f"{column!s} IS {'NOT ' if negative else ''}NULL", args)

        return "FALSE" if not negative else "TRUE", []

    def _build_multi_in(
        self, column: tuple[StatementOrColumn, ...], values: Collection[tuple[Any, ...]], *, negative: bool = False
    ) -> tuple[str, list[Any]]:
        # pylint: disable=consider-using-f-string
        column = tuple(Column(col) if not isinstance(col, Statement) else col for col in column)

        none_multi_values = [value_tuple for value_tuple in values if any(value is None for value in value_tuple)]
        values = [value_tuple for value_tuple in values if all(value is not None for value in value_tuple)]

        args = []

        for stmt in column:
            if isinstance(stmt, Statement):
                args.extend(stmt.args)

        for value_tuple in values:
            for value in value_tuple:
                if not isinstance(value, Statement):
                    args.append(value)
                elif isinstance(value, Statement):
                    args.extend(value.args)

        multi_in_stmt = "({}) {} ({})".format(
            ", ".join(map(str, column)),
            "IN" if not negative else "NOT IN",
            ", ".join(
                [
                    "("
                    + ", ".join(
                        [self.dialect.placeholder if not isinstance(value, Statement) else str(value) for value in value_tuple]
                    )
                    + ")"
                    for value_tuple in values
                ]
            ),
        )

        if not values and not none_multi_values:
            return "FALSE" if not negative else "TRUE", []

        if not none_multi_values:
            return (
                multi_in_stmt,
                args,
            )

        or_stmt = (Or if not negative else And)()

        if values:
            or_stmt.append(Raw(multi_in_stmt, *args))

        for value_tuple in none_multi_values:
            or_stmt.append(And(*[(Eq if not negative else Ne)(col, value) for col, value in zip(column, value_tuple)]))

        return (str(or_stmt), or_stmt.args)

    def __bool__(self) -> bool:
        return True

    def __invert__(self) -> "In":
        """
        Allows using the `~` operator to negate the IN condition, converting it to a NOT IN condition.
        Note: Cannot use ~ operator on NotIn conditions.
        """
        return NotIn(self._column, self._values)


class NotIn(In):
    """
    `NOT IN` condition for checking whether column value is not in given list of values.

    This is a dedicated class for NOT IN conditions, which is equivalent to using In with negative=True.
    It provides a more intuitive API for NOT IN conditions.

    Examples:

    ### Single column

    - Simple NOT IN statement:
        ```python
        # `column` NOT IN (%s, %s, %s)
        NotIn("column", [1, 2, 3])
        "`column` NOT IN (%s, %s, %s)", [1, 2, 3]
        ```
    - Statement NOT IN values
        ```python
        # <statement> NOT IN (%s, %s, %s)
        NotIn(Date("column"), ["2021-01-01", "2021-01-02", "2021-01-03"])
        "Date(`column`) NOT IN (%s, %s, %s)", ["2021-01-01", "2021-01-02", "2021-01-03"]
        ```
    - Value with None works out-of-the-box
        ```python
        NotIn("column", [1, None, 3])
        "`column` NOT IN (%s, %s) AND `column` IS NOT NULL", [1, 3]
        ```

    ### Multi column

    - Against tuples of values:
        ```python
        # (`column1`, `column2`) NOT IN ((%s, %s), (%s, %s), (%s, %s))
        NotIn(("column1", "column2"), [(1, 2), (3, 4), (5, 6)])
        "(`column1`, `column2`) NOT IN ((%s, %s), (%s, %s), (%s, %s))", [1, 2, 3, 4, 5, 6]
        ```
    - Values with `None`:
        ```python
        NotIn(("column1", "column2"), [(1, 2), (3, None), (5, 6)])
        "(`column1`, `column2`) NOT IN ((%s, %s), (%s, %s)) AND NOT (`column1` = %s AND `column2` IS NULL)", [1, 2, 3, 5, 6]
        ```

    ### Subquery NOT IN

    ```python
    NotIn("column", Select("column", table="table", where=Eq("column", 1)))
    ```

    ```python
    NotIn(("column1", "column2"), Select("column1", "column2", table="table", where=Eq("column", 1)))
    ```
    """

    def __init__(
        self,
        column: StatementOrColumn | tuple[StatementOrColumn, ...],
        values: Collection[Any | tuple[Any, ...]] | Select,
        /,
    ) -> None:
        """
        :param column: Column to compare, or tuple of columns for multi-column comparison.
        :param values: Values to compare (list of values, or list of tuples of values for multi-column Not_In).
        """
        super().__init__(column, values, negative=True)

    def __invert__(self) -> NoReturn:
        """
        Allows using the `~` operator to negate the NOT IN condition, converting it to an IN condition.
        Note: Cannot use ~ operator on NotIn conditions.
        """
        raise TypeError("Cannot use ~ operator on NotIn conditions")



"""Module containing definition of executable SQL statements base classes."""

import inspect
from abc import ABC
from typing import Any, Awaitable, Protocol, TypeAlias, TypeVar, cast, overload

from sqlfactory.logger import logger
from sqlfactory.statement import ConditionalStatement, Statement

R_co = TypeVar("R_co", covariant=True)


class HasQueryWithTupleArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with query method that takes arguments as tuple."""

    def query(self, query: str, args: tuple[Any, ...]) -> R_co: ...


class HasExecuteWithTupleArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with execute method that takes arguments as tuple."""

    def execute(self, query: str, args: tuple[Any, ...]) -> R_co: ...


class HasQueryWithArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with query method that takes arguments as multiple arguments."""

    def query(self, query: str, *args: Any) -> R_co: ...


class HasExecuteWithArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with execute method that takes arguments as tuple."""

    def execute(self, query: str, *args: Any) -> R_co: ...


class HasAsyncQueryWithTupleArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with async query method that takes arguments as tuple."""

    async def query(self, query: str, args: tuple[Any, ...]) -> R_co: ...


class HasAsyncExecuteWithTupleArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with async execute method that takes arguments as tuple."""

    async def execute(self, query: str, args: tuple[Any, ...]) -> R_co: ...


class HasAsyncQueryWithArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with async query method that takes arguments as multiple arguments."""

    async def query(self, query: str, *args: Any) -> R_co: ...


class HasAsyncExecuteWithArgs(Protocol[R_co]):
    # pylint: disable=too-few-public-methods, missing-function-docstring  # As this is just a protocol.
    """Protocol defining DB driver with async execute method that takes arguments as tuple."""

    async def execute(self, query: str, *args: Any) -> R_co: ...


HasAsyncQuery: TypeAlias = HasAsyncQueryWithTupleArgs[R_co] | HasAsyncQueryWithArgs[R_co]

HasAsyncExecute: TypeAlias = HasAsyncExecuteWithTupleArgs[R_co] | HasAsyncExecuteWithArgs[R_co]

HasQuery: TypeAlias = HasQueryWithTupleArgs[R_co] | HasQueryWithArgs[R_co]

HasExecute: TypeAlias = HasExecuteWithTupleArgs[R_co] | HasExecuteWithArgs[R_co]

HasQueryOrExecute: TypeAlias = HasQuery[R_co] | HasExecute[R_co]

HasAsyncQueryOrExecute: TypeAlias = HasAsyncQuery[R_co] | HasAsyncExecute[R_co]
MaybeAsyncHasQueryOrExecute: TypeAlias = HasQueryOrExecute[R_co] | HasAsyncQueryOrExecute[R_co]


class ExecutableStatement(Statement, ABC):
    """
    This is the base class for an executable SQL statement that does not have any arguments.

    This class implements the execute() method. When given a database driver (or cursor) with query() or execute()
    methods, which take an SQL statement as the first argument and then a tuple or variadic arguments following SQL
    argument, it can be used to directly execute the SQL statement. This saves some typing by avoiding the manual
    passing of a string statement and arguments to the query.

    DB-API 2.0 drivers/cursors should all work with this implementation, as cursors should have execute() methods with
    the described semantics.

    Even async is supported. As this class does not process the result of the SQL statement in any way, the return
    value of the driver's execute()/query() method is directly returned. That returned value can be awaitable for async
    methods, so you can directly await it.
    """

    @overload
    async def execute(self, trx: HasAsyncQuery[R_co], *args: Any) -> R_co:
        """
        Execute statement on async db driver

        Example:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> await select.execute(async_cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    async def execute(self, trx: HasAsyncExecute[R_co], *args: Any) -> R_co:
        """
        Execute statement on async db driver

        Example:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> await select.execute(async_cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    def execute(self, trx: HasQuery[R_co], *args: Any) -> R_co:
        """
        Execute statement on sync db driver.

        Example:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> select.execute(cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    def execute(self, trx: HasExecute[R_co], *args: Any) -> R_co:
        """
        Execute statement on sync db driver.

        Example:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> select.execute(cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    def execute(self, trx: MaybeAsyncHasQueryOrExecute[R_co], *args: Any) -> R_co | Awaitable[R_co]:
        """
        Execute statement on db driver (db-agnostic, just expects method `query` or `execute` on given driver).
        This is just a shortland for calling driver.execute(str(self), *args).

        Example:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> select.execute(cursor)

        or for async:

        >>> from sqlfactory import Select, Eq
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> select = Select("table").where("column", Eq(42))
        >>> await select.execute(async_cursor)

        :param trx: DB driver with query() or execute() method, which accepts either tuple as arguments,
         or multiple arguments following the query.
        :param args: Arguments to pass to the driver's query/execute method.
        :return: The same as db driver's execute/query method. If driver is async, returns awaitable response.
        """
        if hasattr(trx, "query"):
            call = trx.query
        elif hasattr(trx, "execute"):
            call = trx.execute
        else:
            raise AttributeError("trx must define query() or execute() method.")

        sig = inspect.signature(call)
        if any(p.kind == inspect.Parameter.VAR_POSITIONAL for p in sig.parameters.values()):
            return call(str(self), *self.args, *args)

        return call(str(self), tuple(self.args + list(args)))


class ConditionalExecutableStatement(ExecutableStatement, ConditionalStatement, ABC):
    """
    Mixin that provides conditional execution of the statement (query will be executed only if statement is valid).

    This class is used for example for INSERT statements, to not execute empty INSERT. Or to not execute UPDATE
    if there are no columns to be updated.
    """

    @overload  # type: ignore[override]
    async def execute(self, trx: HasAsyncQuery[R_co], *args: Any) -> R_co | bool:
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.

        Example:

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> await insert.execute(async_cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> await insert.execute(async_cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    async def execute(self, trx: HasAsyncExecute[R_co], *args: Any) -> R_co | bool:
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.

        Example:

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> await insert.execute(async_cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> await insert.execute(async_cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    def execute(self, trx: HasQuery[R_co], *args: Any) -> R_co | bool:
        # pylint: disable=invalid-overridden-method
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.

        Example:

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> insert.execute(cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> insert.execute(cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    @overload
    def execute(self, trx: HasExecute[R_co], *args: Any) -> R_co | bool:
        # pylint: disable=invalid-overridden-method
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.

        Example:

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> insert.execute(cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> insert.execute(cursor)

        :param: Return value from the driver's execute() or query() method.
        """

    def execute(self, trx: MaybeAsyncHasQueryOrExecute[R_co], *args: Any) -> R_co | bool | Awaitable[R_co | bool]:
        # pylint: disable=invalid-overridden-method
        """
        Execute SQL statement using provided db-driver, but only if statement evaluates as True.

        Example:

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> insert.execute(async_cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> insert.execute(async_cursor)

        or for async cursors:

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...  # Your db driver's cursor
        >>> insert = Insert("products")("id", "name").values((1, "test"))
        >>> # Gets executed only if there are any rows to be inserted.
        >>> insert.execute(async_cursor)

        >>> from sqlfactory import Insert
        >>>
        >>> async_cursor: Any = ...
        >>> insert = Insert("products")("id", "name").values()
        >>> # Does not get executed, because there are no rows to be inserted.
        >>> insert.execute(async_cursor)

        :param trx: DB driver with query() or execute() method, which accepts either tuple as arguments,
        :param args: Arguments to pass to the db driver.
        :return:
        """
        if bool(self):
            return cast(R_co, super().execute(trx, *args))

        if inspect.iscoroutinefunction(getattr(trx, "query", getattr(trx, "execute", None))):
            logger.debug("Not executing statement, because it is false.")

            # pylint: disable=import-outside-toplevel
            import asyncio

            fut = asyncio.get_running_loop().create_future()
            fut.set_result(False)
            return fut

        logger.debug("Not executing statement, because it is false.")
        return False
